//// works with ONE fasta sequence

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

typedef int bool;

typedef struct {
   int Col;
   int Row;
} Address;

typedef struct {
	double  Weight;
	bool 	  Match;
	Address Link;
} Record;

//--------------------------------------------------------------------
#define MAX_SEQ_LEN    30000
#define MAX_DESCR_LEN  100
#define LSTR 			  MAX_SEQ_LEN
#define SSTR 			  250 // profile file string
#define DUMMY_WEIGHT   -100000
#define GAP_SYMBOL     '.'

#define END_OF_PROFILE_ID '-'

//--------------------------------------------------------------------

void 	 flush_seq(char *s);
int	 read_prf_file(char *fname);
int	 read_param_file(char *fname);
int	 getIndex(char r);
double align(char *seq);
void	 build_consensus(void);
void	 unshift_seq(char c, char *s);
int	 begin(char *s1, char *s2 );
void 	 set_parameters(void);
void   extract_prf_name(char *src, char *dst);
void   add_seq(char *src, char *dst);
int    index_of_max(double *p, int len);

//--------------------------------------------------------------------

char alphabet[] = "ARNDCQEGHILKMFPSTWYV";


bool prf_free_beg;
bool prf_free_end;
bool seq_free_beg;
bool seq_free_end;

bool show_alignment=0;

bool all_free_beg;
bool all_free_end;

double gap_open_penalty_prf;
double gap_open_penalty_seq;
double deletion_penalty_prf;
double deletion_penalty_seq;

char *consensus;

char *res_seq1=NULL;
char *res_seq2=NULL;

double** profile;
int 		prf_length;



//--------------------------------------------------------------------
void main( int argc,  char *argv[] )
{

	FILE	  *fp;
	char 	  *seq, *descr, *s;
	double  weight;
	char    profile_id[SSTR];



	if(argc!=4) {
		fprintf(stderr, "\nUsage: align  fasta_file prf_file param_file\n");
		return;
	} // end if


   if( !(seq=(char*)malloc(MAX_SEQ_LEN)) ||
	    !(descr=(char*)malloc(MAX_DESCR_LEN)) ||
	    !(s=(char*)malloc(LSTR))
	  ) {
      fprintf(stderr, "\nalloc fail\n");
      return;
   } // end if

	if( !(prf_length=read_prf_file(argv[2])) ) {
		fprintf(stderr, "\ncan't process profile file\n");
		return;
	} // end if

	if( !read_param_file(argv[3]) ) {
		fprintf(stderr, "\ncan't process parameters file\n");
		return;
	} // end if

	set_parameters();

	if(show_alignment)
		build_consensus();


   if( !(fp=fopen(argv[1], "r"))) {
      fprintf(stderr, "\ncan't open %s", argv[1]);
      return;
   } // end if


//   extract_prf_name(argv[2], profile_id);

	strcpy(profile_id, argv[2]);

	flush_seq(seq); flush_seq(descr);


	seq[0] = '\0';

	while( fgets(s, LSTR, fp)) {

		if(s[0]=='>') {
			if(strlen(seq)>0) { // process
				break;
			} // end if strlen
			else {
				sscanf(s+1, "%s", descr); // skip '>'
			}

		} // end if s[0]=='>'
		else {
			add_seq(s, seq);
		} // end else

	} // end while

	fclose(fp);

	if(strlen(seq)>0) { // process

		weight = align(seq);

		printf("Seq: %s  Profile: %s   Score: %8.3lf\n", descr, profile_id, weight);

		if(show_alignment)
			printf("\n%s\n%s\n", res_seq1, res_seq2);

		flush_seq(seq);

	} // end if strlen
	else {
		fprintf(stderr, "\ncan't find fasta sequence in argv[1]\n");
	}


	free(seq); free(descr); free(s);

	if(res_seq1) free(res_seq1);
	if(res_seq2) free(res_seq2);


} // end main()
//--------------------------------------------------------------------
/*

(Peptide) : d119l__-4.2.1.3.1.aln Length: 162
Cons     A      R      N      D      C      Q      E      G      H      I      L      K      M      F      P      S      T      W      Y      V    Gap  Len  ..
000 -0.078 -0.078 -0.078 -0.078 -0.078 -0.078 -0.078 -0.078 -0.078 -0.078 -0.078 -0.078 +1.504 -0.078 -0.078 -0.078 -0.078 -0.078 -0.078 -0.078      1    1
001 -0.237 -0.237 +1.380 -0.237 -0.237 -0.237 -0.237 -0.237 -0.237 -0.237 -0.237 -0.237 -0.237 -0.237 -0.237 +0.557 -0.237 -0.237 -0.237 -0.237      1    1
002 -0.368 -0.368 -0.368 -0.368 -0.368 -0.368 -0.368 +1.034 -0.368 +0.437 +0.223 -0.368 +1.043 -0.368 -0.368 -0.368 -0.368 -0.368 -0.368 -0.368      1    1

*/
int read_prf_file(char *fname)
{

	int    i, len;
	double **pp;
	FILE   *fp;
	char   *s;

   if( !(fp=fopen(fname, "r"))) {
      printf("\ncan't open %s", fname);
      return NULL;
   } // end if

	if( !(s=malloc(SSTR))) {
		fprintf(stderr, "\nError---string alloc fail\n");
		return NULL;
	} // end if

	fgets(s, SSTR, fp);

	if(
		strstr(s, "(Peptide)")!=s  ||
		strstr(s, "Length:")==NULL ||
		sscanf(s, "%*s%*s%*s%*s%d", &len)!=1
	  ) {
      fprintf(stderr, "\ncan't recognize profile file in %s", fname);
      return NULL;
	} // end if

	fgets(s, SSTR, fp); // swallow 'Cons' string


	if( !(pp=(double**)calloc(len, sizeof(double*))) ) {
		fprintf(stderr, "\nError---matrix alloc fail 1\n");
		return NULL;
	} // end if


	for( i=0; i<len; i++) {

		if( !(pp[i] =(double*)calloc(20, sizeof(double))) ) {
			fprintf(stderr, "\nError---matrix alloc fail 2\n");
			return NULL;
		} // end if

		fgets(s, SSTR, fp);

		sscanf(s, "%*s%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf%lf",

			pp[i]+0, pp[i]+1, pp[i]+2, pp[i]+3, pp[i]+4,
			pp[i]+5, pp[i]+6, pp[i]+7, pp[i]+8, pp[i]+9,

			pp[i]+10, pp[i]+11, pp[i]+12, pp[i]+13, pp[i]+14,
			pp[i]+15, pp[i]+16, pp[i]+17, pp[i]+18, pp[i]+19
		);


	} // end for i

	free(s);
	fclose(fp);
	profile = pp;

	return len;

} // end read_prf_file()
//--------------------------------------------------------------------
int read_param_file(char *fname)
{

/*

PRF_GAP_OPEN_PENALTY=4.0
PRF_DELETION_PENALTY=0.4
SEQ_GAP_OPEN_PENALTY=4.0
SEQ_DELETION_PENALTY=0.4
SHOW_ALIGNMENT=NO
PRF_FREE_BEG=NO
PRF_FREE_END=NO
SEQ_FREE_BEG=NO
SEQ_FREE_END=YES

*/


	FILE   *fp;
	char   *s;
	int    i=0;

   if( !(fp=fopen(fname, "r"))) {
      printf("\ncan't open %s", fname);
      return NULL;
   } // end if

	if( !(s=malloc(SSTR))) {
		fprintf(stderr, "\nError---string alloc fail\n");
		return NULL;
	} // end if

	while( fgets(s, SSTR, fp) ) {


		if( begin(s, "PRF_GAP_OPEN_PENALTY")) {
			i += sscanf(s+21, "%lf", &gap_open_penalty_prf);
		}
		else if( begin(s, "PRF_DELETION_PENALTY")) {
			i += sscanf(s+21, "%lf", &deletion_penalty_prf);
		}
		else if( begin(s, "SEQ_GAP_OPEN_PENALTY")) {
			i += sscanf(s+21, "%lf", &gap_open_penalty_seq);
		}
		else if( begin(s, "SEQ_DELETION_PENALTY")) {
			i += sscanf(s+21, "%lf", &deletion_penalty_seq);
		}
		else if( begin(s, "PRF_FREE_BEG")) {
			prf_free_beg = begin(s+13,"YES");
			i++;
		}
		else if( begin(s, "PRF_FREE_END")) {
			prf_free_end = begin(s+13,"YES");
			i++;
		}
		else if( begin(s, "SEQ_FREE_BEG")) {
			seq_free_beg = begin(s+13,"YES");
			i++;
		}
		else if( begin(s, "SEQ_FREE_END")) {
			seq_free_end = begin(s+13,"YES");
			i++;
		}
		else if( begin(s, "SHOW_ALIGNMENT")) {
			show_alignment = begin(s+15,"YES");
			i++;
		}


	} // end while

	free(s);
	fclose(fp);

	return (i==9);


} // end read_param_file()
//--------------------------------------------------------------------
void flush_seq(char *s)
{
	s[0]='\0';
} // end flush_seq()
//--------------------------------------------------------------------
void add_seq(char *src, char *dst)
{

	void cat_char(char *s, char c);

	int i;

	for(i=0; i<strlen(src); i++) {
		if(isalpha(src[i])) {
			cat_char(dst, toupper(src[i]));
		} // end if
	} // end for

} // end add_seq()
//--------------------------------------------------------------------
void cat_char(char *s, char c)
{
	int len;

	len = strlen(s);

	s[len]=c;
	s[len+1]='\0';

} // end cat_char()
//-------------------------------------------------------------------------
double align(char *seq)
{

  int length_1;
  int length_2;

  int maxColIndex;
  int maxRowIndex;


  int init_row = 0;
  int init_col = 0;

  int match_row; // Row_address of the next match
  int match_col; // Col_address of the next match


  // Prepare variables

  Record  *current_row;
  Record  *previous_row;

  Record  best_cell;
  Record  best_cell_1;
  Record  best_cell_2;

  Address **archive_matrix;

  double weights[3];
  int 	work_index;
  double work_weight;
  bool 	work_match;
  int 	row_address=0;
  int 	col_address=0;

  int *seq_index;

  int i, j;

  double aln_weight;

  //-----------------------------------------------------------

  length_1 = prf_length;
  length_2 = strlen(seq);

  maxColIndex = length_1+1;
  maxRowIndex = length_2+1;

  // Prepare aa indecis

// seq_index = new int[maxRowIndex];
	if( !(seq_index=(int*)calloc(maxRowIndex, sizeof(int))) ) {
		fprintf(stderr, "\nError---alloc 1 fail\n");
		return DUMMY_WEIGHT;
	} // end if

   for (i=0; i<maxRowIndex; i++) {
     seq_index[i] = getIndex(seq[i]);
   } // end fir


	////////////////////////////////////
  // Prepare data_structures
  if (prf_free_end) {
   best_cell_1.Weight = DUMMY_WEIGHT;
  }

  if (seq_free_end) {
   best_cell_2.Weight = DUMMY_WEIGHT;
  }

  if (all_free_end) {
   best_cell.Weight = DUMMY_WEIGHT;
  }


	if(
			 !(current_row =(Record*)calloc(maxColIndex, sizeof(Record))) ||
			 !(previous_row=(Record*)calloc(maxColIndex, sizeof(Record))) ||
			 !(archive_matrix =(Address**)calloc(maxRowIndex, sizeof(Address*)))
	  ) {
		fprintf(stderr, "\nError---alloc 2 fail\n");
		return DUMMY_WEIGHT;
	} // end if


	for (i=0; i<maxRowIndex; i++) {
		// archive_matrix[i] = new Address[maxColIndex];
		if( !(archive_matrix[i] =(Address*)calloc(maxColIndex, sizeof(Address))) ) {
			fprintf(stderr, "\nError---alloc 3 fail\n");
			return DUMMY_WEIGHT;
		} // end if
	} // end for

  // Initialization, row #0, cell #0
  // fill records
  current_row[0].Weight   = 0.0;
  current_row[0].Link.Row = init_row;
  current_row[0].Link.Col = init_col;
  current_row[0].Match    = 0;

  // Initialization, row #0, all cells from #1,
  // step by col
	for (j = 1; j<maxColIndex; j++) {

		// fill records
		if ( seq_free_beg || all_free_beg ) {
			current_row[j].Weight = 0.0;
		}
		else {
			current_row[j].Weight = -gap_open_penalty_seq-deletion_penalty_seq*j;
		}

		current_row[j].Link.Row = init_row;
		current_row[j].Link.Col = init_col;
		current_row[j].Match    = 0;

	} // end for j<maxColIndex



  ///////////// Main passage /////////////////////////

  // Step by row
  for (i = 1; i<maxRowIndex; i++) {

   // Copy current_row into previous_row
   for (j = 0; j<maxColIndex; j++) {
	   previous_row[j]   = current_row[j];
   } // end for j<maxColIndex

   // Fill the cell #0 in current row, fill records
   if (prf_free_beg) {
   	current_row[0].Weight = 0.0;
   }
   else {
   	current_row[0].Weight = -gap_open_penalty_prf-deletion_penalty_prf*i;
   }

   current_row[0].Link.Row = init_row;
   current_row[0].Link.Col = init_col;
   current_row[0].Match    = 0;

   // Step by col
   for (j = 1; j<maxColIndex; j++) {

     // Work with records,  set possible weights
     weights[2] = previous_row[j-1].Weight+profile[j-1][seq_index[i-1]];
     if ( (previous_row[j].Link.Row == (i-1)) || (previous_row[j].Match) ) {
	     weights[1] = previous_row[j].Weight-gap_open_penalty_prf-deletion_penalty_prf;
     }
     else {
   	  weights[1] = previous_row[j].Weight-deletion_penalty_prf;
     }

     if ( (current_row[j-1].Link.Col == (j-1)) || (current_row[j-1].Match) ) {
	     weights[0] = current_row[j-1].Weight-gap_open_penalty_seq-deletion_penalty_seq;
     }
     else {
	     weights[0] = current_row[j-1].Weight-deletion_penalty_seq;
     }


     work_index  = index_of_max(weights,3);

     work_weight = weights[work_index];

     if ((all_free_beg) && (work_weight<=0)) {
	     work_weight = 0.0;
	     col_address = init_col;
	     row_address = init_row;
	     work_match = 0;
     }
     else {
     // work with all_records,  set respective links
	     switch (work_index) {
	      case 2:if (previous_row[j-1].Match) {
	              row_address = i-1;
	              col_address = j-1;
	             }
	             else {
	              row_address = previous_row[j-1].Link.Row;
	              col_address = previous_row[j-1].Link.Col;
	             }
	             work_match = 1;
	             break;
	      case 1:if (previous_row[j].Match) {
	              row_address = i-1;
	              col_address = j;
	             }
	             else {
	              row_address = previous_row[j].Link.Row;
	              col_address = previous_row[j].Link.Col;
	             }
	             work_match = 0;
	             break;
	      case 0:if (current_row[j-1].Match) {
	              row_address = i;
	              col_address = j-1;
	             }
	             else {
	              row_address = current_row[j-1].Link.Row;
	              col_address = current_row[j-1].Link.Col;
	             }
	             work_match = 0;
	             break;
	     } // end switch
     } // end else

     // Work with records, fill active_cell
     current_row[j].Weight   = work_weight;
     current_row[j].Match    = work_match;
     current_row[j].Link.Row = row_address;
     current_row[j].Link.Col = col_address;

     // Fill archive_cell
     archive_matrix[i][j].Row = row_address;
     archive_matrix[i][j].Col = col_address;


     // work with best_cells
     if (all_free_end && work_index==2 && work_weight>best_cell.Weight ) {

        best_cell.Weight   = work_weight;
        best_cell.Match    = 1;
        best_cell.Link.Row = i;
        best_cell.Link.Col = j;

     } // end if all_free_end

     if (prf_free_end && j==maxColIndex-1 && work_index==2 && work_weight>best_cell_1.Weight ) {

         best_cell_1.Weight   = work_weight;
         best_cell_1.Match    = 1;
         best_cell_1.Link.Row = i;
         best_cell_1.Link.Col = j;

     } // end prf_free_end

     if (seq_free_end && i==maxRowIndex-1 && work_index==2 && work_weight>best_cell_2.Weight) {

         best_cell_2.Weight   = work_weight;
         best_cell_2.Match    = 1;
         best_cell_2.Link.Row = i;
         best_cell_2.Link.Col = j;

     } // end seq_free_end

   } // end for j<maxColIndex
  } // end for i<maxRowIndex


   if (aln_weight>DUMMY_WEIGHT/2) {

	   if (all_free_end) {
		   aln_weight = best_cell.Weight;
	   } // end if
	   else if (prf_free_end) {
		   aln_weight = best_cell_1.Weight;
	   } // end elseif
	   else if (seq_free_end) {
		   aln_weight = best_cell_2.Weight; ////////////////////////
	   } // end elseif

   } // end if
	else {
  		aln_weight = current_row[maxColIndex-1].Weight;
	} // end else

  // Reverse


  if (show_alignment) {

	if(
			!(res_seq1=(char*)calloc(MAX_SEQ_LEN, sizeof(char))) ||
			!(res_seq2=(char*)calloc(MAX_SEQ_LEN, sizeof(char)))

	) {
		fprintf(stderr, "\nError---alloc fail\n");
		return DUMMY_WEIGHT;
	} // end if

   j = maxColIndex-1;
   i = maxRowIndex-1;

   if (aln_weight>DUMMY_WEIGHT/2) {

	   if (all_free_end) {

		   match_row  = best_cell.Link.Row;
		   match_col  = best_cell.Link.Col;

		   flush_seq(res_seq1);
		   flush_seq(res_seq2);

		   i++; j++;

	   } // end if
	   else if (prf_free_end) {


		   match_row = best_cell_1.Link.Row;
		   match_col = best_cell_1.Link.Col;

		   flush_seq(res_seq1);
		   flush_seq(res_seq2);

			i++; j++;
	   } // end elseif
	   else if (seq_free_end) {

		   match_row = best_cell_2.Link.Row;
		   match_col = best_cell_2.Link.Col;

		   flush_seq(res_seq1);
		   flush_seq(res_seq2);

	      i++; j++;

	   } // end elseif
	   else {


		   match_row = archive_matrix[i][j].Row;
		   match_col = archive_matrix[i][j].Col;
		   // If last cell in archive matrix contain match,
		   // then init align sequences by last chars of start sequences...
		   if (current_row[maxColIndex-1].Match) {

  			   res_seq1[0] = consensus[j-1];
  			   res_seq2[0] = seq[i-1];

		   } // end if
		   // ...else init align sequences as...
		   else {

			   flush_seq(res_seq1);
			   flush_seq(res_seq2);

			   i++;  j++;

		   } // end else

	   } // end else


	   while (1) {

		   i--;  j--;


		   if ( (i!=match_row) || (j!=match_col) ) {

			   while (i>match_row) {

					unshift_seq(GAP_SYMBOL, res_seq1);
					unshift_seq(seq[i-1],   res_seq2);
			      i--;

			   } // end while
			   while (j>match_col) {


					unshift_seq(GAP_SYMBOL, res_seq2);
					unshift_seq(consensus[j-1],   res_seq1);
			      j--;

			   } // end while
		   } // end if

		   // If not the cell #0 then add match to alignment...
		   if ((j!=0) && (i!=0)) {
				unshift_seq(seq[i-1], res_seq2);
				unshift_seq(consensus[j-1],   res_seq1);
		   }
		   // ...else end
		   else {
			   while (i>0) {
					unshift_seq(GAP_SYMBOL, res_seq1);
					unshift_seq(seq[i-1],   res_seq2);
			      i--;
			   } // end while
			   while (j>0) {
					unshift_seq(GAP_SYMBOL, res_seq2);
					unshift_seq(consensus[j-1],   res_seq1);
			      j--;
			   } // end while
			   break;
		   } // end else
		   // Take address of the previous match
		   match_row = archive_matrix[i][j].Row;
		   match_col = archive_matrix[i][j].Col;
	   } // end while

   } // end if
   else {

		flush_seq(res_seq1);
		flush_seq(res_seq2);

   } // end else


  } // end show_alignment

  ///////////////// Free the memory ////////////////////

  free(seq_index); free(current_row); free(previous_row);

  for (i=0; i<maxRowIndex; i++) {
    free(archive_matrix[i]);
  } // end for

  free(archive_matrix);

  return aln_weight;

} // end align()
//--------------------------------------------------------------------
int getIndex(char r)
{
	int i;

	for ( i=0; i<20; i++)
		if (alphabet[i]==toupper(r))
			return i;

	return -1;
} // end getIndex
//--------------------------------------------------------------------
void build_consensus(void)
{

	int i, j;
  double best_score;
  int		best_idx;


	if( !(consensus=(char*)calloc(prf_length+1, sizeof(char))) ) {
		fprintf(stderr, "\nError---alloc 1 fail\n");
		return;
	} // end if

	flush_seq(consensus);

	for(i=0; i<prf_length; i++) {
		best_score = DUMMY_WEIGHT;
		best_idx   = -1;
		for(j=0; j<20; j++) {
			if( profile[i][j]>best_score) {
				best_score = profile[i][j];
				best_idx   = j;
			} // end if
		} // end for j

		consensus[i] = tolower(alphabet[best_idx]);

	} // end for i

	consensus[prf_length] = '\0';

} // end build_consensus
//--------------------------------------------------------------------
void unshift_seq(char c, char *s)
{

	int i;

	i = strlen(s);

	s[i+1]='\0';

	while(i) {
		s[i]=s[i-1];
		i--;
	} // end while

	s[0]=c;

} // end unshift_seq()
//--------------------------------------------------------------------
int begin(char *s1, char *s2 )
{
	return (strstr(s1, s2)==s1);
} // end begin();
//--------------------------------------------------------------------
void set_parameters(void)
{

	if( seq_free_end && prf_free_end ) {
		all_free_end = 1;
		seq_free_end = prf_free_end = 0;
	} // end if
	else {
		all_free_end = 0;
	}

	if( seq_free_beg && prf_free_beg ) {
		all_free_beg = 1;
		seq_free_beg = prf_free_beg = 0;
	} // end if
	else {
		all_free_beg = 0;
	}

} // end set_parameters()
//--------------------------------------------------------------------
int index_of_max(double *p, int len)
{

	int 	k, index = -1;
	double maxp = DUMMY_WEIGHT;

	for(k=0; k<len; k++) {
		if( p[k]>=maxp) {
			index = k;
			maxp = p[k];
		} // end if
	} // end for

	return index;

} // end index_of_max()
//--------------------------------------------------------------------
void extract_prf_name(char *src, char *dst)
{

	char *end;

	strcpy(dst, src);

	if( (end=strchr(dst, END_OF_PROFILE_ID))!=NULL) {
		*end = '\0';
	} // end if

} // end extract_prf_name()
//--------------------------------------------------------------------
